#include "../core/utils/utils.h"
#include "../core/debug/evemon.h"
#include "../core/debug/exception.h"
#include "../core/threads/event.h"
#include "../core/instancers/instancer.h"
#include "../core/test/test.h"
#include "../core/resources/resources.h"

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#ifdef WIN32
#define EOL	"\r\n"
#else
#define EOL "\n"
#endif

/* invalid_snippet_exception exception class */
class invalid_snippet_exception : public exception::base
{
public:
	virtual std::string to_string(void) const override
	{
		return "Invalid character in snippet file!";
	}
};

/* snippet class */
class snippet
{
public:
	snippet(const std::string& message)
	{
		set(message);
	}

	snippet(storage::buffer data)
	{
		// add a '\0' to force string termination
		data.append<char>('\0');

		// set string
		set(data.ptr<char>(0));
	}

	const std::string& get(void) const
	{
		return this->m_message;
	}

private:
	void set(const std::string& message)
	{
		// throw exception if it has non-ascii character in it
		for (size_t i = 0; i < message.length(); i++)
			if (!isvalid(message.at(i)))
				throw_exception(invalid_snippet_exception);

		// copy string
		this->m_message = message;

		// shorten
		if (this->m_message.length() > 77)
			this->m_message = this->m_message.substr(0, 77) + "...";
	}

	bool isvalid(const char c) const
	{
		return c >= 0x20 && c <= 0x7e;
	}

private:
	std::string m_message;
};

// exit flag
thread::event g_exit;

void sigfunc(int sigval)
{
	// trigger exit signal
	g_exit.trigger();
}

// main program
int main(int argc, char** argv)
{
	if (argc != 2)
	{
		fprintf(stderr, "Usage: %s <file>" EOL, argv[0]);

		return 1;
	}

	// exit event on interrupt
	signal(SIGINT, &sigfunc);
	signal(SIGTERM, &sigfunc);

	try
	{
		// set null resource
		resources::set_null_resource(std::make_shared<const snippet>("Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum."));

		// create resource
		resources::async_proxy<snippet> res(std::string("file://") + std::string(argv[1]));

		// print every 100 ms as long as we don't quit
		g_exit.do_until([&](void)
			{
				auto message = res->get();

				printf("\r%s", message.c_str());

				for (size_t i = message.length(); i < 83; i++)
					printf(" ");

				std::this_thread::sleep_for(std::chrono::duration<double, std::ratio<1>>(0.1));
			});

		printf(EOL);

		// return successfuly
		return 0;
	}
	catch (exception::base& e)
	{
		printf("error: %s" EOL, e.to_string().c_str());

		_critical("%s", e.to_string().c_str());
	}
	catch (std::exception& e)
	{
		printf("error: %s" EOL, e.what());

		_critical("%s", e.what());
	}
	catch (...)
	{
		printf("unknown error!" EOL);

		_critical("unknown error!");
	}

	// signal error
	return 1;
}
